@@ -1,14 +1,34 @@
+#
+# Copyright © 2025 Starlake AI (https://starlake.ai)
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+#
+
 import os
 
 from datetime import timedelta
 
 import logging
 
-from typing import Any, Dict, Sequence, Union
+from typing import Any, Dict, Optional, Sequence, Union
 
-from ai.starlake.job import StarlakePreLoadStrategy, StarlakeSparkConfig, StarlakeExecutionEnvironment
+from ai.starlake.dataset import StarlakeDataset
 
-from ai.starlake.airflow import StarlakeAirflowJob
+from ai.starlake.job import StarlakePreLoadStrategy, StarlakeSparkConfig, StarlakeExecutionEnvironment, TaskType
+
+from ai.starlake.airflow import StarlakeAirflowJob, StarlakeDatasetMixin
+
+from ai.starlake.airflow.bash import StarlakeBashOperator
 
 from airflow.exceptions import AirflowException
 
@@ -19,7 +39,7 @@ from airflow.operators.bash import BashOperator
 from airflow.providers.google.cloud.hooks.cloud_run import CloudRunHook
 from airflow.providers.google.cloud.operators.cloud_run import  CloudRunExecuteJobOperator
 
-from airflow.sensors.base import BaseSensorOperator
+from airflow.sensors.base import BaseSensorOperator, PokeReturnValue
 from airflow.sensors.bash import BashSensor
 
 from airflow.utils.context import Context
@@ -35,8 +55,8 @@ class StarlakeAirflowCloudRunJob(StarlakeAirflowJob):
     """Airflow Starlake Cloud Run Job."""
     def __init__(
             self, 
-            filename: str, 
-            module_name: str,
+            filename: str = None, 
+            module_name: str = None,
             pre_load_strategy: Union[StarlakePreLoadStrategy, str, None]=None,
             project_id: str=None,
             cloud_run_job_name: str=None,
@@ -75,19 +95,33 @@ class StarlakeAirflowCloudRunJob(StarlakeAirflowJob):
         """
         return StarlakeExecutionEnvironment.CLOUD_RUN
 
-    def sl_job(self, task_id: str, arguments: list, spark_config: StarlakeSparkConfig=None, **kwargs) -> BaseOperator:
+    def sl_job(self, task_id: str, arguments: list, spark_config: StarlakeSparkConfig=None, dataset: Optional[Union[StarlakeDataset, str]]=None, task_type: Optional[TaskType] = None, **kwargs) -> BaseOperator:
         """Overrides StarlakeAirflowJob.sl_job()
         Generate the Airflow task that will run the starlake command.
         
         Args:
             task_id (str): The required task id.
             arguments (list): The required arguments of the starlake command to run.
+            spark_config (Optional[StarlakeSparkConfig], optional): The optional spark configuration. Defaults to None.
+            dataset (Optional[Union[StarlakeDataset, str]], optional): The optional dataset to materialize. Defaults to None.
+            task_type (Optional[TaskType], optional): The optional task type. Defaults to None.
             
         Returns:
             BaseOperator: The Airflow task.
         """
         kwargs.update({'pool': kwargs.get('pool', self.pool)})
         kwargs.update({'retry_delay': timedelta(seconds=self.retry_delay_in_seconds)})
+        if task_type is not None and (task_type == TaskType.LOAD or task_type == TaskType.TRANSFORM):
+            arguments = [] if not arguments else arguments
+            params: dict = kwargs.get('params', dict())
+            cron = params.get('cron_expr', params.get('cron', None))
+            params.update({'cron': cron})
+            kwargs.update({'params': params})
+            tmp_arguments = []
+            tmp_arguments.append("--scheduledDate")
+            tmp_arguments.append("\'{{sl_scheduled_date(params.cron, ts_as_datetime(data_interval_end | ts))}}\'")
+            command = arguments.pop(0)
+            arguments = [command] + tmp_arguments + arguments
         command = f'^{self.separator}^' + self.separator.join(arguments)
         if self.cloud_run_async: # asynchronous job
             with TaskGroup(group_id=f'{task_id}_wait') as task_completion_sensors:
@@ -107,6 +141,8 @@ class StarlakeAirflowCloudRunJob(StarlakeAirflowJob):
                     check_completion_id = task_id + '_check_completion'
                     completion_sensor = GCloudRunJobCompletionSensor(
                         task_id=check_completion_id,
+                        dataset=dataset if self.retry_on_failure else None,
+                        source=self.source,
                         project_id=self.project_id,
                         cloud_run_job_region=self.cloud_run_job_region,
                         source_task_id=job_task.task_id,
@@ -126,20 +162,22 @@ class StarlakeAirflowCloudRunJob(StarlakeAirflowJob):
                             bash_command=f"""
                             set -e
                             bash -c '
-                            {bash_command}
+                            {bash_command.replace("'", '"')}
                             return_code=$?
 
                             # Push the return code to XCom
                             echo $return_code
 
                             # Exit with the captured return code if non-zero
-                            if [ $return_code -ne 0 ]; then
-                                exit $return_code
-                            fi
+                            # if [ $return_code -ne 0 ]; then
+                            #     exit $return_code
+                            # fi
                             '
                             """
-                        job_status = BashOperator(
+                        job_status = StarlakeBashOperator(
                             task_id=get_completion_status_id,
+                            dataset=dataset,
+                            source=self.source,
                             bash_command=bash_command,
                             **kwargs
                         )
@@ -155,6 +193,8 @@ class StarlakeAirflowCloudRunJob(StarlakeAirflowJob):
                     job_overrides = {"container_overrides": [container_overrides]}
                     job_task = CloudRunJobOperator(
                         task_id=task_id,
+                        dataset=None,
+                        source=self.source,
                         project_id=self.project_id,
                         job_name=self.cloud_run_job_name,
                         region=self.cloud_run_job_region,
@@ -166,6 +206,8 @@ class StarlakeAirflowCloudRunJob(StarlakeAirflowJob):
                     check_completion_id = task_id + '_check_completion'
                     completion_sensor = CloudRunJobCompletionSensor(
                         task_id=check_completion_id,
+                        dataset=dataset,
+                        source=self.source,
                         source_task_id=job_task.task_id,
                         impersonation_chain=self.impersonate_service_account,
                         **kwargs
@@ -173,7 +215,7 @@ class StarlakeAirflowCloudRunJob(StarlakeAirflowJob):
 
                     job_task >> completion_sensor
 
-                return task_completion_sensors
+            return task_completion_sensors
 
         else: # synchronous job
             if self.use_gcloud:
@@ -187,21 +229,25 @@ class StarlakeAirflowCloudRunJob(StarlakeAirflowJob):
                     bash_command=f"""
                     set -e
                     bash -c '
-                    {bash_command}
+                    {bash_command.replace("'", '"')}
                     return_code=$?
 
                     # Push the return code to XCom
                     echo $return_code
 
                     # Exit with the captured return code if non-zero
-                    if [ $return_code -ne 0 ]; then
-                        exit $return_code
-                    fi
+                    # if [ $return_code -ne 0 ]; then
+                    #     exit $return_code
+                    # fi
                     '
                     """
-                return BashOperator(
+                kwargs.pop('do_xcom_push', None)
+                return StarlakeBashOperator(
                     task_id=task_id,
+                    dataset=dataset,
+                    source=self.source,
                     bash_command=bash_command,
+                    do_xcom_push=True,
                     **kwargs
                 )
             else:
@@ -214,6 +260,8 @@ class StarlakeAirflowCloudRunJob(StarlakeAirflowJob):
                 job_overrides = {"container_overrides": [container_overrides]}
                 return CloudRunJobOperator(
                     task_id=task_id,
+                    dataset=dataset,
+                    source=self.source,
                     project_id=self.project_id,
                     job_name=self.cloud_run_job_name,
                     region=self.cloud_run_job_region,
@@ -223,11 +271,22 @@ class StarlakeAirflowCloudRunJob(StarlakeAirflowJob):
                     **kwargs
                 )
 
-class GCloudRunJobCompletionSensor(BashSensor):
+class GCloudRunJobCompletionSensor(StarlakeDatasetMixin, BashSensor):
     '''
     This sensor checks the completion of a cloud run job using gcloud.
     '''
-    def __init__(self, *, project_id: str, cloud_run_job_region: str, source_task_id: str, retry_on_failure: bool=None, impersonate_service_account: str=None, **kwargs) -> None:
+    def __init__(self, 
+                 *, 
+                 task_id: str, 
+                 dataset: Optional[Union[StarlakeDataset, str]],
+                 source: Optional[str],
+                 project_id: str, 
+                 cloud_run_job_region: str, 
+                 source_task_id: str, 
+                 retry_on_failure: bool=None, 
+                 impersonate_service_account: str=None, 
+                 **kwargs
+        ) -> None:
         if retry_on_failure:
             kwargs.update({'retry_exit_code': 2})
             bash_command = (
@@ -249,42 +308,53 @@ class GCloudRunJobCompletionSensor(BashSensor):
         else:
             bash_command=(f"value=`gcloud beta run jobs executions describe {{{{task_instance.xcom_pull(key='return_value', task_ids='{source_task_id}')}}}}  --region {cloud_run_job_region} --project {project_id} --format='value(status.completionTime, status.cancelledCounts)' {impersonate_service_account}| sed 's/[[:blank:]]//g'`; test -n \"$value\"")
 
-        # if kwargs.get('do_xcom_push', False):
-        #     bash_command=f"""
-        #     set -e
-        #     bash -c '
-        #     {bash_command}
-        #     return_code=$?
+        if kwargs.get('do_xcom_push', False) and retry_on_failure:
+            bash_command=f"""
+            set -e
+            bash -c '
+            {bash_command.replace("'", '"')}
+            return_code=$?
 
-        #     # Push the return code to XCom
-        #     echo $return_code
+            # Push the return code to XCom
+            echo $return_code
 
-        #     # Exit with the captured return code if non-zero
-        #     if [ $return_code -ne 0 ]; then
-        #         exit $return_code
-        #     fi
-        #     '
-        #     """
+            # Exit with the captured return code if non-zero
+            # if [ $return_code -ne 0 ]; then
+            #     exit $return_code
+            # fi
+            '
+            """
         super().__init__(
+            task_id=task_id,
+            dataset=dataset,
+            source=source,
             bash_command=bash_command,
             mode="reschedule",
             **kwargs
         )
 
-class CloudRunJobOperator(CloudRunExecuteJobOperator):
+class CloudRunJobOperator(StarlakeDatasetMixin, CloudRunExecuteJobOperator):
     """
     This extends official CloudRunExecuteJobOperator in order to implement asynchronous job.
     """
 
     def __init__(
         self,
+        task_id: str, 
+        dataset: Optional[Union[StarlakeDataset, str]],
+        source: Optional[str],
         mode: CloudRunMode = CloudRunMode.SYNC,
         gcp_conn_id: str = "google_cloud_default",
         impersonation_chain: Union[str, Sequence[str], None] = None,
         **kwargs,
     ):
         super().__init__(  # type: ignore
-            gcp_conn_id=gcp_conn_id, impersonation_chain=impersonation_chain, **kwargs
+            task_id=task_id,
+            dataset=dataset,
+            source=source,
+            gcp_conn_id=gcp_conn_id, 
+            impersonation_chain=impersonation_chain, 
+            **kwargs
         )
         self.mode = mode
 
@@ -310,21 +380,37 @@ class CloudRunJobOperator(CloudRunExecuteJobOperator):
             return self.operation.operation.name
 
         else:
-            return super(CloudRunJobOperator, self).execute(context)
+            try:
+                job = super(CloudRunJobOperator, self).execute(context)
+                if self.do_xcom_push:
+                    self.xcom_push(context, key="job", value=job)
+                return True
+            except Exception as e:
+                logger.exception(msg=f"Task {self.task_id} has failed")
+                return False
 
-class CloudRunJobCompletionSensor(BaseSensorOperator):
+class CloudRunJobCompletionSensor(StarlakeDatasetMixin, BaseSensorOperator):
 
     template_fields = ("gcp_conn_id", "impersonation_chain")
 
     def __init__(
         self,
         *,
+        task_id: str, 
+        dataset: Optional[Union[StarlakeDataset, str]],
+        source: Optional[str],
         source_task_id: str,
         gcp_conn_id: str = "google_cloud_default",
         impersonation_chain: Union[str, Sequence[str], None] = None,
         **kwargs,
     ):
-        super().__init__(mode="reschedule", **kwargs)
+        super().__init__(
+            task_id=task_id,
+            dataset=dataset,
+            source=source,
+            mode="reschedule", 
+            **kwargs
+        )
         self.source_task_id = source_task_id
         self.gcp_conn_id = gcp_conn_id
         self.impersonation_chain = impersonation_chain
@@ -343,8 +429,14 @@ class CloudRunJobCompletionSensor(BaseSensorOperator):
             # An operation can only have one of those two combinations: if it is failed, then
             # the error field will be populated, else, then the response field will be.
             if operation.error.SerializeToString():
-                raise AirflowException(
-                    f"{operation.error.message} [{operation.error.code}]"
-                )
-            return True
-        return False
+                if self.do_xcom_push:
+                    self.log.error(
+                        f"{operation.error.message} [{operation.error.code}]"
+                    )
+                    return PokeReturnValue(True, False)
+                else:
+                    raise AirflowException(
+                        f"{operation.error.message} [{operation.error.code}]"
+                    )
+            return PokeReturnValue(True, True)
+        return PokeReturnValue(False, False)
